



runbenders --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
evaluate_xhat --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
runph --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
runef --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
phsolverserver --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
scenariotreeserver --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
computeconf --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
test.pyomo --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pyomo --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
